#include "gbaspc.h"
#include "apu.h"

extern const u8 iplRom[64];

u32 apuTimerCounter IN_EWRAM;
u8 apuTimer0Enabled IN_EWRAM;
u8 apuTimer1Enabled IN_EWRAM;
u8 apuTimer2Enabled IN_EWRAM;
u8 apuTimer0Counter IN_EWRAM;
u8 apuTimer1Counter IN_EWRAM;
u8 apuTimer2Counter IN_EWRAM;
u8 apuTimerSkipCounter IN_EWRAM;

static u8 apuShowRom IN_EWRAM;

void ApuResetTimers() {
	apuTimerCounter = 0;
	apuTimer0Counter = 0;
	apuTimer0Enabled = 0;
	apuTimer1Counter = 0;
	apuTimer1Enabled = 0;
	apuTimer2Counter = 0;
	apuTimer2Enabled = 0;
	apuTimerSkipCounter = 0;
}

void ApuWriteControlByte(u8 byte) {
	if ((byte & 0x1) && !apuTimer0Enabled) {
		apuTimer0Counter = 0;
		APU_MEM[APU_COUNTER0] = 0;
	}
	if ((byte & 0x2) && !apuTimer1Enabled) {
		apuTimer1Counter = 0;
		APU_MEM[APU_COUNTER1] = 0;
	}
	if ((byte & 0x4) && !apuTimer2Enabled) {
		apuTimer2Counter = 0;
		APU_MEM[APU_COUNTER2] = 0;
	}

	apuTimer0Enabled = byte & 0x1;
	apuTimer1Enabled = (byte >> 1) & 0x1;
	apuTimer2Enabled = (byte >> 2) & 0x1;

	if (byte & 0x10) {
		// Clear port 0 and 1
		APU_MEM[0xF4] = 0;
		APU_MEM[0xF5] = 0;
	}
	if (byte & 0x20) {
		// Clear port 0 and 1
		APU_MEM[0xF6] = 0;
		APU_MEM[0xF7] = 0;
	}

	if (byte & 0x80) {
		if (!apuShowRom) {
			apuShowRom = 1;
			for (int i=0; i<=0x3F; i++) APU_MEM[0xFFC0 + i] = iplRom[i];
		}
	} else {
		if (apuShowRom) {
			apuShowRom = 0;
			for (int i=0; i<=0x3F; i++) APU_MEM[0xFFC0 + i] = APU_EXTRA_MEM[i];
		}
	}
}

void ApuPrepareStateAfterReload() {
	ApuResetTimers();

	apuShowRom = 1;
	ApuWriteControlByte(APU_MEM[APU_CONTROL_REG]);
}

// Call to update the timers
// int cycles - the number of cycles that have passed by since last timer update
void ApuUpdateTimers(u32 cycles) {
	u32 orig = apuTimerCounter;
	apuTimerCounter += cycles;
	// Hack for wrapround on 32 bits
	if (orig > apuTimerCounter) orig = 0;

	if (apuTimer0Enabled) {
		// Updates is the number of elapsed timer counter overflows since last call.
		u32 updates = (apuTimerCounter / t64Cycles) - (orig / t64Cycles);
		while (updates--) {
			apuTimer0Counter++;
			if (apuTimer0Counter == APU_MEM[APU_TIMER0]) {
				apuTimer0Counter = 0;
				APU_MEM[APU_COUNTER0] = (APU_MEM[APU_COUNTER0] + 1) & 0xf;
			}
		}
	}
	if (apuTimer1Enabled) {
		u32 updates = (apuTimerCounter / t8Cycles) - (orig / t8Cycles);
		while (updates--) {
			apuTimer1Counter++;
			if (apuTimer1Counter == APU_MEM[APU_TIMER0]) {
				apuTimer1Counter = 0;
				APU_MEM[APU_COUNTER1] = (APU_MEM[APU_COUNTER1] + 1) & 0xf;
			}
		}
	}
	if (apuTimer2Enabled) {
		u32 updates = (apuTimerCounter / t8Cycles) - (orig / t8Cycles);
		while (updates--) {
			apuTimer2Counter++;
			if (apuTimer2Counter == APU_MEM[APU_TIMER0]) {
				apuTimer2Counter = 0;
				APU_MEM[APU_COUNTER2] = (APU_MEM[APU_COUNTER2] + 1) & 0xf;
			}
		}
	}
}

u32 ApuReadCounter(u32 cyles, u32 address) {
	if ((APU_MEM[address] & 0xf) == 0) {
		apuTimerSkipCounter++;
		if (apuTimerSkipCounter > 5) {
			apuTimerSkipCounter = 0;

			if (apuTimer0Enabled) {
				
			}
			
			// Skip an arbitrary amount of CPU time baby!
			return 0;
		}
	} else {
		apuTimerSkipCounter = 0;
	}
	APU_MEM[address] = 0;
	return 0;
}

////////////////////////////////////////////////////////////////////////////
// Fast lookup table into memory
////////////////////////////////////////////////////////////////////////////

extern u32 MemWriteDoNothing VAR_IN_IWRAM;
extern u32 MemWriteApuControl VAR_IN_IWRAM;
extern u32 MemWriteDspAddress VAR_IN_IWRAM;
extern u32 MemWriteDspData VAR_IN_IWRAM;

#include "memwritelut.h"

extern u32 MemReadDoNothing VAR_IN_IWRAM;
extern u32 MemReadCounter VAR_IN_IWRAM;

#include "memreadlut.h"
