	.SECTION .iwram, "ax", %progbits
	.ARM
	.ALIGN

.GLOBAL		DspMixOneChannel
/* r0 is number of samples to mix
   r1 is pointer to the channel
   r2 is the address of the mix buffer

	DspChannel is as follows -   
0	u32 lowSpeed;
4	u32 count;
8	s8 *curSample;
12	s8 *loopEnd;
16	s8 *loopStart;
20	u16 highSpeed;
22	u16 pad;
24	s8 volume;
25	u8 active;
26	u8 flags;	// looping
*/
DspMixOneChannel:
	stmdb sp!,{r4-r12}

	stmdb sp!,{r1}
	mov r8, r2		@ address of mix buffer
	mov r2, r0		@ number of samples to mix
	mov r0, r1		@ base of structure
	ldrsb r3,[r0, #24]	@ volume
	ldr   r4,[r0, #0]	@ low speed
	ldrh  r5,[r0, #20]	@ high speed
	ldr   r6,[r0, #4]	@ count
	ldr   r7,[r0, #8]	@ cur sample position
	ldr   r9,[r0, #12]	@ loop end
	ldrb r12,[r0,#26] @ flags
	mov r10,#1		@ sample active

	cmp r12,#0
	beq MixLoopbackNoLoop
	b MixLooping

/* r0 and r1 and temporaries
   r2 = number of samples to mix
   r3 = volume of sample
   r4 = low speed
   r5 = high speed
   r6 = count
   r7 = pointer to 16 bit signed sample
   r8 = mix buffer address
   r9 = loop end
   r10 = sample active*/
MixLoopbackNoLoop:
	ldrsh	r0,[r7]
	adds	r6,r6,r4
	addcs	r7,r7,#2
	add		r7,r7,r5,lsl #1
	cmp		r7,r9
	bge		MixNoLoopOff	@ break out of loop
	ldr		r1,[r8]
	mla		r1,r3,r0,r1
	str		r1,[r8],#4
	subs	r2,r2,#1
	bne		MixLoopbackNoLoop
	b MixDone
MixNoLoopOff:
	mov		r6,#0
	mov		r10,#0		@ set sample to inactive
	b MixDone

/* r11 = loop length */
MixLooping:
	ldr r11,[r0,#16]		@ loop start
	sub r11,r9,r11
MixLoopbackLooping:
	ldrsh	r0,[r7]
	adds	r6,r6,r4
	addcs	r7,r7,#2
	add		r7,r7,r5,lsl #1
	cmp		r7,r9
	subge	r7,r7,r11	@ loopback to beginning
	movge	r10, #0
	ldr		r1,[r8]
	mla		r1,r3,r0,r1
	str		r1,[r8],#4
	subs	r2,r2,#1
	bne		MixLoopbackLooping
	b MixDone

MixDone:
	/* Update the structure */
	ldmia	sp!,{r1}
	str		r6, [r1,#4]	@ count
	str		r7, [r1,#8] @ current sample pos
	strb	r10, [r1,#25] @ sample active or not

	/* Return */		
	ldmia	sp!,{r4-r12}
	bx lr

.GLOBAL		DspDownSample
	/*
	Arguments:
	r0 - source buffer (s32)
	r1 - dest buffer (s8)
	r2 - length
	*/
DspDownSample:
	mov		r2, r2, lsr #1
loopbackDownSample:
	ldr		r3,[r0],#4
	mov		r3,r3,lsr#14
	strb	r3,[r1],#1
	ldr		r3,[r0],#4
	mov		r3,r3,lsr#14
	strb	r3,[r1],#1
	subs	r2,r2,#1
	bne loopbackDownSample
	bx lr
