#include "gbaspc.h"
#include "sprintf.h"

volatile s32 cyclesToExecute;
u8 soundBuffer[MIXBUFSIZE * 2] DATA_ALIGNED_EWRAM;
u32 curSoundBuffer;
void Isr(void);

void CpuFastClear(u32 data, void *dest, u32 elements) {
	CpuFastSet(&data, dest, (1<<24) | elements);
}

void SetupSound() {
	CpuFastClear(0, soundBuffer, MIXBUFSIZE / 2);

	curSoundBuffer = 0;
	DMA[1].src = soundBuffer;
	DMA[1].dst = (void *)0x040000a0;
	DMA[1].count = 1;
	DMA[1].control = DMA_DSTUNCH | DMA_SRCINC | DMA_REPEAT | DMA_U32 | DMA_SPECIAL;

	SETSNDRES(1);

	DSOUNDCTRL = DSOUNDCTRL_ARESET | DSOUNDCTRL_ATIMER(0) | DSOUNDCTRL_AR | DSOUNDCTRL_AL |
				 DSOUNDCTRL_DMG100 | DSOUNDCTRL_A100;

	// Set the frequency
	// sample rate = 16780000 / (65536 - TM0D)
	// TM0D = - (16780000 - 65536(rate)) / rate
	// 22020Hz = 64774 .. this is 367 samples per frame
	// Samples per frame = 22020Hz / 60fps
//	*(vu16*)REG_TM0CNT_L = -(16780000 - 65536*MIXRATE) / MIXRATE;

	TIMER[0].count = 0x10000 - 532; // 31536Hz playback
	TIMER[0].control = TIMER_16MHZ | TIMER_ENABLE;

	TIMER[1].count = 0x10000 - MIXBUFSIZE;
	TIMER[1].control = TIMER_CASCADE | TIMER_IRQ | TIMER_ENABLE;

	SNDSTAT = SNDSTAT_ENABLE;
}

extern const u8 fontData[];

void DrawChar(int sx, int sy, char c) {
	u16 *vram = VRAM + ((sy * 240 + sx) >> 1);
	const u16 *tmpData = (u16*)&(fontData[(c-32)<<6]);
	int y;
	for (y = 0; y < 8; y++) {
		vram[0] = *tmpData++;
		vram[1] = *tmpData++;
		vram[2] = *tmpData++;
		vram[3] = *tmpData++;
		vram += 120;
	}
}

int DrawString(int sx,int sy,char *s) {
	int orgx = sx;
	while (*s != 0) {
		if (*s == '\n') {
			sy += 8;
			sx = orgx;
		} else {
			DrawChar(sx, sy, *s);
			sx += 8;
		}
		s++;
	}
	return sx;
}

void PrintDisasm(int sx,int sy,int num,u16 pc) {
	char buf[100];
	DebugGetSPCState(buf);
	DrawString(sx, sy, buf);
	sy += 24;
	while (num--) {
		pc += DebugGetOpcodeDesc(buf, &(APU_MEM[pc]), true);
		int x = DrawString(sx, sy, buf);
		while (x < 240) {
			DrawChar(x, sy, ' ');
			x += 8;
		}
		sy += 8;
	}
}

void PrintDspState() {
	char c[70];

//	u8 *sampleBase = APU_MEM + (DSP_MEM[DSP_DIR] << 8);
	for (u32 i = 0; i < 8; i++) {
//		u16 *lookupPointer = (u16*)(sampleBase + (DSP_MEM[(i << 4) + DSP_SRC] << 2));
		u16 freq = *((u16*)&(DSP_MEM[(i << 4) + DSP_PITCH_L]));

		sprintf(c, "%d %d %.2d %.4x %3d", channels[i].active, (DSP_MEM[DSP_KON] >> i) & 1,
			DSP_MEM[(i << 4) + DSP_SRC], freq, DSP_MEM[(i << 4) + DSP_ENVX]);

		DrawString(0, i*8, c);
	}
}

void PrintChannelStates() {
	char c[70];
	for (int i=0; i<8; i++) {
		sprintf(c,"%d,%d-%.7x,%.7x,%.7x",channels[i].active, channels[i].flags,
			(u32)channels[i].sample, (u32)channels[i].sampleEnd, 
			(u32)channels[i].loopStart);
		DrawString(0, i*8, c);
	}
}

void LoadSpc(const u8 *spc) {
	ApuReset();
	DspReset();

	APU_STATE[1] = spc[0x27]<<24; // A
	APU_STATE[2] = spc[0x28]<<24; // X
	APU_STATE[3] = spc[0x29]<<24; // Y
	SetStateFromRawPSW(APU_STATE, spc[0x2A]);
	APU_STATE[4] = 0x100 | spc[0x2B]; // SP
	APU_STATE[7] = spc[0x25] | (spc[0x26] << 8); // PC

	for (int i=0; i<=0xffff; i++) APU_MEM[i] = spc[0x100 + i];
	for (int i=0; i<=0x7f; i++) {
		DSP_MEM[i] = spc[0x10100 + i];
		DSP_MEM[0x80|i] = spc[0x10100 + i];
	}
	for (int i=0; i<=0x3f; i++) APU_EXTRA_MEM[i] = spc[0x101c0 + i];

	ApuPrepareStateAfterReload();
	DspPrepareStateAfterReload();	
}

// Slow version of wait for vblank
void WaitForBlankNoIrq() {
	while (LCD_Y > 159);
	while (LCD_Y < 160);
}

void WaitForBlank() {
	__asm volatile("mov r2, #0; swi 0x05" ::: "r0", "r1", "r2", "r3");
}

const u32 *getFileTable() {
	const u8* start = ((const u8 *)0x08000000);
	for (;;) {
		if (start[0] == 123 && start[1] == 154 && start[2] == 143 && start[3] == 36 &&
			start[4] == 201 && start[5] == 243 && start[6] == 9   && start[7] == 255) {
			break;
		}
		start += 256;
	}
	return (u32*)(start + 8);
}

const u8 *getFile(const u32 *fileTable, u32 index) {
	if (fileTable[index] == 0) return 0;
	return ((const u8 *)0x08000000) + fileTable[index];
}

void showMenu() {
	const u32 *fileTable = getFileTable();

	u32 startFile = 0;
	for (;;){
		for (;;) {
			WaitForBlank();
			CpuFastClear(0, VRAM, 240*160/4);
			for (u32 i=startFile; i<startFile+20; i++) {
				const u8 *file = getFile(fileTable, i);
				if (file == 0) break;
				for (int j=0; j<30; j++) {
					if (file[j] == 0) break;
					DrawChar(j * 8, (i - startFile) * 8, file[j]);
				}
			}
			while (JOY & JOY_A) {
				if (!(JOY & JOY_DOWN)) {
					startFile++;
					if (getFile(fileTable, startFile) == 0) startFile--;
					for (int i=0; i<4; i++) WaitForBlank();
					break;
				}
				if (!(JOY & JOY_UP)) {
					if (startFile != 0) startFile--;
					for (int i=0; i<4; i++) WaitForBlank();
					break;
				}
			}
			if (!(JOY & JOY_A)) break;
		}

		// Start up sound
		SetupSound();
		// Load the spc
		LoadSpc(getFile(fileTable, startFile) + 64);

		cyclesToExecute = 0;
		for (;;) {
			for (int i = 0; i < spcUpdatesPerSec; i++) {
				if ((i & 0xff) == 0) PrintDspState();
				while (cyclesToExecute < 0);
				ApuExecute(spcCyclesPerUpdate);
				ApuUpdateTimers(spcCyclesPerUpdate);
				cyclesToExecute -= spcCyclesPerUpdate;
				if (!(JOY & JOY_B)) {
					goto done;
				}
			}
		}

done:
		DspReset();
	}
}

extern const u8 const modes[256];
extern const u8 const modesToBytes[];

void testInstructionSet() {
	LoadSpc(getFile(getFileTable(), 0) + 64);

	// Do some serious debugging here baby
	for (int i = 0; i < 256; i++) {
		int mode = modes[i];
		int bytes = modesToBytes[mode];
		APU_MEM[0] = 0x80;

		for (int j=0; j<=0xff; j++) {
			char c[5];
			sprintf(c,"%-5d",j);
			DrawString(0, 0, c);
			for (int k=0; k<=0xff; k++) {
				//					sprintf(c,"%-5d",k);
				//					DrawString(5*8, 0, c);

				APU_STATE[7] = 0x200;

				APU_MEM[APU_STATE[7]] = i;
				APU_MEM[APU_STATE[7] + 1] = j;
				APU_MEM[APU_STATE[7] + 2] = k;

				if (j == 0 && k ==0) {
					PrintDisasm(0, 0, 19, APU_STATE[7]);
				}

				ApuDebug();

				if (bytes < 3) break;
			}
			if (bytes < 2) break;
		}

		/*			while (JOY & JOY_A);
		for (int i = 0; i < 8; i++) WaitForBlank();*/
	}
}

int main(void) {
	// Enable interrupts, vblank interrupt enabled
	SET_MASTER_ISR(Isr);
	LCDSTAT = LCDSTAT_VBLIRQ;
	INTMASK = INT_VBLANK | INT_TIMER(0) | INT_TIMER(1);
	INTENABLE = 1;

	// Setup mode 4, with white as color 255
	WaitForBlank();
	LCDMODE = 4 | LCDMODE_BG2;
	PALRAM[0xff] = 0x7fff;

	// Nice menu
	showMenu();

	// Start up sound
	SetupSound();

	{
		LoadSpc(getFile(getFileTable(), 1) + 64);

		u16 lastPC = APU_STATE[7];

		for (;;) {
			PrintDisasm(0,0,19,lastPC);
			lastPC = APU_STATE[7];

			ApuDebug();

			if (APU_STATE[7] > lastPC + 3 || APU_STATE[7] < lastPC) {
				lastPC = APU_STATE[7];
			}

			while (JOY & JOY_A);
			for (int i = 0; i < 8; i++) WaitForBlank();
		}
	}

	for(;;) WaitForBlank();

	return 0;
}
