////////////////////////////////////////////////////////////////////////////
// Definitions
////////////////////////////////////////////////////////////////////////////

#define APU_CONTROL_REG		0xF1

#define APU_TIMER0			0xFA
#define APU_TIMER1			0xFB
#define APU_TIMER2			0xFC

#define APU_COUNTER0		0xFD
#define APU_COUNTER1		0xFE
#define APU_COUNTER2		0xFF

// Flags will be represented implicitly in the ARM status
// A reg will be r5, X will be r6, Y will be r7, SP will be r8
// Ram base will be r9
// Direct page pointer will be r10, it's value will be either 0, or 0x100

// PC will be stored in R11, however it's value will be incorrect until the end of a block.
// At the end of a block, it will be corrected to point to the new location it should be at.

// mov pc, lr will be the easy way of breaking out of execution
// # of cycles to execute will be on the stack directly above that

#define APU_REG_A ARM_R5
#define APU_REG_X ARM_R6
#define APU_REG_Y ARM_R7
#define APU_REG_SP ARM_R8
#define APU_RAMBASE ARM_R9
#define APU_DP ARM_R10
#define APU_REG_PC	ARM_R11

// Cycles per second
//#define spcCyclesPerSec 2048000
#define spcCyclesPerSec 1024000
#define spcUpdatesPerSec 2048
#define spcCyclesPerUpdate (spcCyclesPerSec / spcUpdatesPerSec)

// 64Khz timer clock divisor
#define t64Cycles (spcCyclesPerSec / 64000)
// 8Khz timer clock divisor
#define t8Cycles (t64Cycles * 8)

// Externs so we can access these for debugging
extern u8 *APU_MEM;
extern u8 APU_EXTRA_MEM[0x40];

// APU_STATE[0] = CPSR, 1 = A, 2 = X, 3 = Y, 4 = SP, 5 = RAMBASE, 6 = DP
// 7 = PC, 8 = # of cycles to execute
extern u32 APU_STATE[16];

u8 MakeRawPSWFromState(u32 state[16]);
void SetStateFromRawPSW(u32 state[16], u8 psw);

void ApuReset();
void ApuPrepareStateAfterReload();
void ApuDebug();
void ApuExecute(int cycles);
void ApuUpdateTimers(u32 cycles);
