#include "stdafx.h"

const string DataOperationNames[16] = {
	"AND","EOR","SUB","RSB","ADD","ADC","SBC","RSC","TST","TEQ","CMP","CMN","ORR","MOV","BIC","MVN"
};

// Always is an empty string, that is not a mistake ;)
const string CondNames[16] = {
	"EQ","NE","CS","CC","MI","PL","VS","VC","HI","LS","GE","LT","GT","LE","","NV"
};

void GenerateInstructionSetHeader(FILE *outFile) {
	// Generate data-operation commands
	fprintf(outFile,"#include \"armDefs.h\"\n\n");
	fprintf(outFile,"// Automatically generated list of ARM operations\n\n");
	for (int cond=0; cond<=14; cond++) {
		for (int flags=0; flags<=1; flags++) {
			for (int op=0; op<=15; op++) {
				// Immediate mode operations
				fprintf(outFile,"#define %s%s%s_I(rd, rn, rot, immed) armEmitDataOp(%d, armBuildDataOpcode(%d, %d, %d), (rn), (rd), armBuildShiftImm((rot), (immed)))\n",
					DataOperationNames[op].c_str(),
					flags == 0 ? "" : "S",
					CondNames[cond].c_str(),
					cond,
					op, 1, flags);

				// Register modified by immediate
				fprintf(outFile,"#define %s%s%s_RI(rd, rn, rm, type, shift) armEmitDataOp(%d, armBuildDataOpcode(%d, %d, %d), (rn), (rd), armBuildShiftRegImm((shift), (type), (rm)))\n",
					DataOperationNames[op].c_str(),
					flags == 0 ? "" : "S",
					CondNames[cond].c_str(),
					cond,
					op, 0, flags);

				// Register modified by register
				fprintf(outFile,"#define %s%s%s_R(rd, rn, rm, type, rs) armEmitDataOp(%d, armBuildDataOpcode(%d, %d, %d), (rn), (rd), armBuildShiftRegReg((rs), (type), (rm)))\n",
					DataOperationNames[op].c_str(),
					flags == 0 ? "" : "S",
					CondNames[cond].c_str(),
					cond,
					op, 0, flags);
			}
		}

		// B and BL instruction
		for (int link=0; link<=1; link++) {
			fprintf(outFile,"#define B%s%s(address) armEmitBranch(%d, %d, (address))\n",
				link == 0 ? "" : "L",
				CondNames[cond].c_str(),
				cond,
				link);
		}

		// BX instruction
		fprintf(outFile,"#define BX%s(reg) armEmitBranchX(%d, (reg))\n",
			CondNames[cond].c_str(),
			cond);

		// Data load/store instructions
		for (int byte = 0; byte <= 1; byte++) {
			for (int post = 0; post <= 1; post++) {
				for (int add = 0; add <= 1; add++) {
					for (int writeback = 0; writeback <= 1; writeback++) {
						for (int load = 0; load <= 1; load++) {
							// Load/store word/byte immediate
							fprintf(outFile, "#define %s%s%s%s%s_I(rn, rd, imm) armEmitLoadStoreWord(%d, %d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreWordAddressingImmediate((imm)))\n",
								load == 0 ? "STR" : "LDR",
								byte == 0 ? "" : "B",
								post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
								add == 1 ? "" : "N",
								CondNames[cond].c_str(),
								cond, 0, post, add, byte, writeback, load);

							// Load/store word/byte register
							fprintf(outFile, "#define %s%s%s%s%s_R(rn, rd, rm) armEmitLoadStoreWord(%d, %d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreWordAddressingRegister((rm)))\n",
								load == 0 ? "STR" : "LDR",
								byte == 0 ? "" : "B",
								post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
								add == 1 ? "" : "N",
								CondNames[cond].c_str(),
								cond, 1, post, add, byte, writeback, load);

							// Load/store word/byte scaled register
							fprintf(outFile, "#define %s%s%s%s%s_RS(rn, rd, amount, type, rm) armEmitLoadStoreWord(%d, %d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreWordAddressingScaledRegister((amount), (type), (rm)))\n",
								load == 0 ? "STR" : "LDR",
								byte == 0 ? "" : "B",
								post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
								add == 1 ? "" : "N",
								CondNames[cond].c_str(),
								cond, 1, post, add, byte, writeback, load);

							if (byte == 1) {
								// Load/store word register
								fprintf(outFile, "#define %sSB%s%s%s_I(rn, rd, imm) armEmitLoadStoreHalfword(%d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreHalfwordAddressingImmediate((imm)), armEmitSignedByteType)\n",
									load == 0 ? "STR" : "LDR",
									post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
									add == 1 ? "" : "N",
									CondNames[cond].c_str(),
									cond, post, add, 1, writeback, load);

								fprintf(outFile, "#define %sSB%s%s%s_R(rn, rd, rm) armEmitLoadStoreHalfword(%d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreHalfwordAddressingRegister((rm)), armEmitSignedByteType)\n",
									load == 0 ? "STR" : "LDR",
									post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
									add == 1 ? "" : "N",
									CondNames[cond].c_str(),
									cond, post, add, 0, writeback, load);
							} else {
								fprintf(outFile, "#define %sH%s%s%s_I(rn, rd, imm) armEmitLoadStoreHalfword(%d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreHalfwordAddressingImmediate((imm)), armEmitHalfwordType)\n",
									load == 0 ? "STR" : "LDR",
									post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
									add == 1 ? "" : "N",
									CondNames[cond].c_str(),
									cond, post, add, 1, writeback, load);

								fprintf(outFile, "#define %sH%s%s%s_R(rn, rd, rm) armEmitLoadStoreHalfword(%d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreHalfwordAddressingRegister((rm)), armEmitHalfwordType)\n",
									load == 0 ? "STR" : "LDR",
									post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
									add == 1 ? "" : "N",
									CondNames[cond].c_str(),
									cond, post, add, 0, writeback, load);

								fprintf(outFile, "#define %sSH%s%s%s_I(rn, rd, imm) armEmitLoadStoreHalfword(%d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreHalfwordAddressingImmediate((imm)), armEmitSignedHalfwordType)\n",
									load == 0 ? "STR" : "LDR",
									post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
									add == 1 ? "" : "N",
									CondNames[cond].c_str(),
									cond, post, add, 1, writeback, load);

								fprintf(outFile, "#define %sSH%s%s%s_R(rn, rd, rm) armEmitLoadStoreHalfword(%d, %d, %d, %d, %d, %d, (rn), (rd), armEmitLoadStoreHalfwordAddressingRegister((rm)), armEmitSignedHalfwordType)\n",
									load == 0 ? "STR" : "LDR",
									post == 0 ? (writeback == 0 ? "P" : "T") : (writeback == 0 ? "" : "W"),
									add == 1 ? "" : "N",
									CondNames[cond].c_str(),
									cond, post, add, 0, writeback, load);
							}
						}
					}
				}
			}
		}

		// Multiply instructions
		for (int flags=0; flags<=1; flags++) {
			fprintf(outFile,"#define MLA%s%s(rd, rm, rs, rn) armEmitMultiplyAccumulate(%d, %d, (rd), (rn), (rs), (rm))\n",
            	flags == 0 ? "": "S",
				CondNames[cond].c_str(),
				cond,
				flags);

			fprintf(outFile,"#define MUL%s%s(rd, rm, rs) armEmitMultiply(%d, %d, (rd), (rs), (rm))\n",
            	flags == 0 ? "": "S",
				CondNames[cond].c_str(),
				cond,
				flags);
		}

		// Move to/from status register
		for (int reg=0; reg<=1; reg++) {
			fprintf(outFile,"#define MSR%s%s_I(rot, imm) armEmitRegToPSRImm(%d, %d, armEmitPSRFlags, (rot), (imm))\n",
				reg == 0 ? "" : "S",
				CondNames[cond].c_str(),
				cond,
				reg);

			fprintf(outFile,"#define MSR%s%s_R(rm) armEmitRegToPSRReg(%d, %d, armEmitPSRFlags, (rm))\n",
				reg == 0 ? "" : "S",
				CondNames[cond].c_str(),
				cond,
				reg);

			fprintf(outFile,"#define MRS%s%s(rm) armEmitPSRToReg(%d, %d, (rm))\n",
				reg == 0 ? "" : "S",
				CondNames[cond].c_str(),
				cond,
				reg);
		}

		// Load/store multiple
		for (int p=0; p<=1; p++) {
			for (int up=0; up<=1; up++) {
				for (int writeback=0; writeback<=1; writeback++) {
					for (int load=0; load<=1; load++) {
 						fprintf(outFile,"#define %s%s%s%s(rn, regs) armEmitLoadStoreMultiple(%d, %d, %d, %d, %d, (rn), (regs))\n",
							load == 0 ? "STM" : "LDM",
							p == 0 ? (up == 0 ? "DA" : "IA") : (up == 0 ? "DB" : "IB"),
							writeback == 0 ? "" : "W",
							CondNames[cond].c_str(),
							cond,
							p,
							up,
							writeback,
							load);
					}
				}
			}
		}

		fprintf(outFile, "#define SWI%s(immed) armEmitSoftwareInterrupt(%d, (immed))\n",
			CondNames[cond].c_str(),
			cond);
	}
}